#include <stdint.h> // for intptr_t
#include <string.h>
#include <stdbool.h>
#include "ev3api.h"

#if defined(BUILD_MODULE)
#include "module_cfg.h"
#else
#include "kernel_cfg.h"
#endif

#include "syssvc/serial.h"
#include "action.h"
#include "app.h"

#define DEBUG
#ifdef DEBUG
#define _debug(x) (x)
#else
#define _debug(x)
#endif

typedef enum Human {
	undetected = 0,
	detected
} Human;

typedef enum Splay {
	stop = 0,
	splay
} Splay;

typedef enum CoverMode {
	cover_close = 0,
	cover_open
} CoverMode;

static Splay splay_mode = stop;
static CoverMode cover_mode = cover_close;
static Human human_mode = undetected;

static FLGPTN flg_sessalet = 0;	//TODO イベントフラグに変更
static FLGPTN flg_splay = 0;
static FLGPTN flg_seated = 0;

static int   	bt_cmd = 0;     // Bluetooth
static FILE*	bt = NULL;     // Bluetooth

// スプレーボタン（着座時のみ有効）
static void spray_clicked_handler(intptr_t unused) {
	switch (splay_mode) {
	case stop:
		if (flg_seated == SEAT_OFF) {	//着座してない時は洗浄しない
			return;
		}
		// スプレー開始
		flg_splay = SPLAY_ON;
		act_tsk(SPLAY_TASK);
		break;
	case splay:
		// スプレー停止
		flg_splay = SPLAY_OFF;
		break;
	default:
		break;
	}
}

// Flash（流水音）ボタンの操作。いつでも流れる、時間がきたら自動的に止まる
// ここで操作するのは流すのみ。止まるはフラッシュタスク内で制御
static void button_clicked_handler(intptr_t button) {
	switch (button) {
	case RIGHT_BUTTON:
		act_tsk(FLASH_TASK);
		break;
	default:
		break;
	}
}

// 人検知
// 超音波センサーで読み取った距離で人がいるかどうかを判断する
int human_check(void) {
	int result = undetected;
	int distance = ev3_ultrasonic_sensor_get_distance(human_sensor);
	if (distance < SONAR_DIST) {
		result = detected;
	}
	return result;
}

void human_check_task(intptr_t unused) {
	// 最初少し読み捨てた方がよさそう
	(void)human_check();
	dly_tsk(100);
	(void)human_check();
	dly_tsk(100);
	while (true) {
		int result = human_check();
		switch (result) {
		case detected:
			flg_sessalet = DETECTED;
			human_mode = detected;
			break;
		case undetected:
			flg_sessalet = UNDETECTED;
			human_mode = detected;
			break;
		default:
			break;
		}
		tslp_tsk(1000);
	}
}

// Flashタスク流水音の開始と指定時間後の停止
void flash_task(intptr_t unused) {
	startFlash(FLASH_TIME);
}

// カラーセンサーの値で着座判定
int seated_check(void) {
	const int reflect = ev3_color_sensor_get_reflect(seated_sensor);
	if ((SEATED_MIN < reflect)  && (reflect < SEATED_MAX)) {
		return detected;
	} else {
		return undetected;
	}
}

void seated_check_task(intptr_t unused) {
	while (true) {
		int result = seated_check();
		if (result == detected) {
			flg_seated = SEAT_ON;
		} else {
			flg_seated = SEAT_OFF;
		}
		dly_tsk(500);
	}
}

void splay_task(intptr_t unused) {
	int counter = 0;
	while (true) {
		if ((flg_splay == SPLAY_ON) && (flg_seated == SEAT_ON) && (splay_mode == stop)) { // スプレーON受信 -> スプレー開始
			splay_mode = splay;
			startSpray();
			break;
		} else if (flg_splay == SPLAY_OFF) { // スプレーOFF -> タスク終了
			stopSpray();
			splay_mode = stop;
			wup_tsk(SEATED_TASK);
			ext_tsk();
		} else if (flg_seated == SEAT_OFF) { // 起立検出 -> スプレー中だったら中止してタスク終了
			if (splay_mode != stop) {	//スプレー停止
				stopSpray();
				splay_mode = stop;
			}
			wup_tsk(SEATED_TASK);
			ext_tsk();
		} else {
			dly_tsk(5000);
		}
	}
	while (true) {
		if (flg_splay == SPLAY_OFF) { //スプレー停止
			stopSpray();
			splay_mode = stop;
			break;
		} else {
			counter++;
			if (counter > SPLAY_TIMER) {
				//タイムアウトになったらスプレー停止
				stopSpray();
				splay_mode = stop;
				break;
			}
		}
		dly_tsk(1000);
	}
	if (splay_mode != stop) {
		stopSpray();
		splay_mode = stop;
	}
	wup_tsk(SEATED_TASK);
	ext_tsk();
}


//着座タスク
void seated_task(intptr_t unused) {
	// 洗浄待機中
	ev3_button_set_on_clicked(LEFT_BUTTON, spray_clicked_handler, LEFT_BUTTON);

	while (true) {
		if (flg_seated == SEAT_OFF) {
			wup_tsk(MAIN_TASK);
			ext_tsk();
			break;
		}
		else if ((flg_seated == SEAT_ON) && (flg_splay == SPLAY_ON)) {
			wup_tsk(SPLAY_TASK);
			slp_tsk();
		}
	}
	ext_tsk();
}

// 初期化
void initialize(void) {
	ev3_button_set_on_clicked(RIGHT_BUTTON, button_clicked_handler, RIGHT_BUTTON);

	ev3_sensor_config(human_sensor, ULTRASONIC_SENSOR);
	ev3_sensor_config(seated_sensor, COLOR_SENSOR);
	ev3_sensor_config(gyro_sensor, GYRO_SENSOR);

	ev3_motor_config(splay_motor, LARGE_MOTOR);
	ev3_motor_config(flash_motor, LARGE_MOTOR);
	ev3_motor_config(cover_motor, LARGE_MOTOR);
	ev3_motor_config(nozzle_motor, LARGE_MOTOR);
	ev3_motor_reset_counts(cover_motor);

	//水の音量
	ev3_speaker_set_volume(10);
	// Open Bluetooth file(ログ用)
	bt = ev3_serial_open_file(EV3_SERIAL_BT);
	assert(bt != NULL);
	act_tsk(BT_TASK);

	act_tsk(HUMAN_CHECK_TASK);
	act_tsk(SEATED_CHECK_TASK);

	ev3_lcd_fill_rect(0, 0, EV3_LCD_WIDTH, EV3_LCD_HEIGHT, EV3_LCD_WHITE);
	ev3_lcd_draw_string("initialize ok", 0, CALIB_FONT_HEIGHT * 1);
}

void main_task(intptr_t unused) {
	char msg[100];
	initialize();
	dly_tsk(500);	// ハードの制御完了前に戻ってくる可能性があるので、初期化後少し待つ
	ev3_led_set_color(LED_ORANGE);	// 初期化完了を画面に表示

	ev3_lcd_fill_rect(0, 0, EV3_LCD_WIDTH, EV3_LCD_HEIGHT, EV3_LCD_WHITE);
	ev3_lcd_draw_string("main task L315", 0, CALIB_FONT_HEIGHT * 1);

	// 最初は着座OFFモード
	while (true) {
		// 人検出か着座検出を待つ。5秒間たったらふたを閉める
		bool timeout_flg = false;
		int timeout_counter = 0;
		while (timeout_flg == 0) {	//タイムアウトがきたら1に変わっているはず
			tslp_tsk(1000);
			if ((flg_sessalet == DETECTED) && (flg_seated == SEAT_ON)) { //着座検出 -> 着座状態へ遷移
				memset(msg, 0x00, sizeof(msg));
				sprintf(msg, "--- act seated task, from main task ---\r\n");
				(void)serial_wri_dat(SIO_PORT_BT, msg, strlen(msg));
				act_tsk(SEATED_TASK);
				slp_tsk();
			} else if (flg_sessalet == DETECTED) {
				if (cover_mode == cover_close) {
					startOpen();
					cover_mode = cover_open;
				}
			} else if (flg_sessalet == UNDETECTED) {
				if (cover_mode == cover_open) {
					timeout_counter++;
					if (timeout_counter > 5) {
						timeout_flg = true;
					}
				}
			} else if (flg_sessalet == SEAT_OFF) {

			}
		}
		if (timeout_flg == true) {
			timeout_flg = false;
			timeout_counter = 0;
			startClose();
			dly_tsk(3000);	// フタが閉まるまで待つ（パカパカするの防止）
			cover_mode = cover_close;
		}

	}
	ext_tsk();
}

// Bluetooth通信用のTask（Bluetooth経由でログ出力用）
void bt_task(intptr_t unused) {
	while (true) {
		const uint8_t c = fgetc(bt);
		switch (c) {
		case '1':
			bt_cmd = 1;
			break;
		default:
			break;
		}
		fputc(c, bt);
	}
}
